@@ -1,5 +1,7 @@
 /*
  * Copyright (C) 2003 Sun Microsystems, Inc.
+ * Copyright (C) 2004 Red Hat, Inc.
+ * Copyright (C) 2004 Novell, Inc.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -18,6 +20,7 @@
  *
  * Authors:
  *      Mark McLoughlin <mark@skynet.ie>
+ *      Federico Mena Quintero <federico@ximian.com>
  *
  *
  *   The screen polling code is based on XUpdateScanner from
@@ -40,6 +43,9 @@
 #ifdef HAVE_XSHM
 #include <X11/extensions/XShm.h>
 #endif
+#ifdef HAVE_XDAMAGE
+#include <X11/extensions/Xdamage.h>
+#endif
 
 #include "vino-util.h"
 
@@ -52,9 +58,13 @@ typedef struct { int dummy; } XShmSegmentInfo;
 
 struct _VinoFBPrivate
 {
+  Display         *xdisplay;
   GdkScreen       *screen;
   GdkWindow       *root_window;
+
   XImage          *fb_image;
+  XShmSegmentInfo  fb_image_x_shm_info;
+  Pixmap           fb_pixmap;
 
   XImage          *scanline;
   XShmSegmentInfo  scanline_x_shm_info;
@@ -67,7 +77,17 @@ struct _VinoFBPrivate
 
   guint            update_timeout;
 
+#ifdef HAVE_XDAMAGE
+  Damage           xdamage;
+  int              xdamage_notify_event;
+  XserverRegion    xdamage_region;
+  GC               xdamage_copy_gc;
+#endif
+
   guint            use_x_shm : 1;
+  guint            use_xdamage : 1;
+
+  guint            fb_image_is_x_shm_segment : 1;
   guint            scanline_is_x_shm_segment : 1;
   guint            tile_is_x_shm_segment : 1;
 };
@@ -90,6 +110,12 @@ enum
 static void vino_fb_init_from_screen (VinoFB    *vfb,
 				      GdkScreen *screen);
 
+#ifdef HAVE_XDAMAGE
+static GdkFilterReturn vino_fb_xdamage_event_filter (GdkXEvent *xevent,
+						     GdkEvent  *event,
+						     VinoFB    *vfb);
+#endif
+
 static gpointer parent_class;
 static guint    signals [LAST_SIGNAL] = { 0 };
 
@@ -117,13 +143,11 @@ vino_fb_get_image (VinoFB          *vfb,
 		   int              width,
 		   int              height)
 {
-  Display  *xdisplay;
-  Drawable  xdrawable;
-  int       error;
+  Drawable xdrawable;
+  int      error;
 
   g_assert (vfb != NULL && drawable != NULL && image != NULL && x_shm_info != NULL);
 
-  xdisplay  = GDK_DISPLAY_XDISPLAY (gdk_screen_get_display (vfb->priv->screen));
   xdrawable = GDK_DRAWABLE_XID (drawable);
 
   gdk_error_trap_push ();
@@ -131,14 +155,17 @@ vino_fb_get_image (VinoFB          *vfb,
 #ifdef HAVE_XSHM  
   if (is_x_shm_segment && image->width == width && image->height && height)
     {
-      XShmGetImage (xdisplay, xdrawable,
-		    image, x, y,
+      XShmGetImage (vfb->priv->xdisplay,
+		    xdrawable,
+		    image,
+		    x, y,
 		    AllPlanes);
     }
   else
 #endif /* HAVE_XSHM */
     {
-      XGetSubImage (xdisplay, xdrawable,
+      XGetSubImage (vfb->priv->xdisplay,
+		    xdrawable,
 		    x, y, width, height,
 		    AllPlanes, ZPixmap,
 		    image, 0, 0);
@@ -146,10 +173,10 @@ vino_fb_get_image (VinoFB          *vfb,
 
   if ((error = gdk_error_trap_pop ()))
     {
-#if 0
+#ifdef G_ENABLE_DEBUG
       char error_text [64];
 
-      XGetErrorText (xdisplay, error, error_text, 63);
+      XGetErrorText (vfb->priv->xdisplay, error, error_text, 63);
 
       g_warning ("Received a '%s' X Window System error while copying a tile",
 		 error_text);
@@ -169,17 +196,18 @@ static void
 vino_fb_destroy_image (VinoFB          *vfb,
 		       XImage          *image,
 		       XShmSegmentInfo *x_shm_info,
-		       gboolean         is_x_shm_segment)
+		       gboolean         is_x_shm_segment,
+		       gboolean         is_attached)
 {
 #ifdef HAVE_XSHM
   if (is_x_shm_segment)
     {
+      if (is_attached)
+	XShmDetach (vfb->priv->xdisplay, x_shm_info);
+
       if (x_shm_info->shmaddr != (char *)-1)
 	shmdt (x_shm_info->shmaddr);
       x_shm_info->shmaddr = (char *)-1;
-
-      if (x_shm_info->shmid != -1)
-	shmctl (x_shm_info->shmid, IPC_RMID, 0);
       x_shm_info->shmid = -1;
     }
 #endif /* HAVE_XSHM */
@@ -192,22 +220,21 @@ static gboolean
 vino_fb_create_image (VinoFB           *vfb,
 		      XImage          **image,
 		      XShmSegmentInfo  *x_shm_info,
+		      gboolean          must_use_x_shm,
 		      int               width,
 		      int               height,
-		      int               bits_per_pixel)
+		      int               depth)
 {
-  Display *xdisplay;
-  int      n_screen;
+  int n_screen;
 
-  xdisplay = GDK_DISPLAY_XDISPLAY (gdk_screen_get_display (vfb->priv->screen));
   n_screen = gdk_screen_get_number (vfb->priv->screen);
 
 #ifdef HAVE_XSHM
   if (vfb->priv->use_x_shm)
     {
-      *image = XShmCreateImage (xdisplay,
-			       DefaultVisual (xdisplay, n_screen),
-			       bits_per_pixel,
+      *image = XShmCreateImage (vfb->priv->xdisplay,
+			       DefaultVisual (vfb->priv->xdisplay, n_screen),
+			       depth,
 			       ZPixmap,
 			       NULL,
 			       x_shm_info,
@@ -231,8 +258,8 @@ vino_fb_create_image (VinoFB           *vfb,
 
       gdk_error_trap_push ();
 
-      XShmAttach (xdisplay, x_shm_info);
-      XSync (xdisplay, False);
+      XShmAttach (vfb->priv->xdisplay, x_shm_info);
+      XSync (vfb->priv->xdisplay, False);
       
       if (gdk_error_trap_pop ())
 	goto x_shm_error;
@@ -244,25 +271,26 @@ vino_fb_create_image (VinoFB           *vfb,
     x_shm_error:
       vfb->priv->use_x_shm = FALSE;
 
-      vino_fb_destroy_image (vfb, *image, x_shm_info, TRUE);
+      vino_fb_destroy_image (vfb, *image, x_shm_info, TRUE, FALSE);
       *image = NULL;
       
-      return vino_fb_create_image (vfb, image, x_shm_info,
-				   width, height, bits_per_pixel);
+      return vino_fb_create_image (vfb, image, x_shm_info, FALSE,
+				   width, height, depth);
     }
-  else
 #endif /* HAVE_XSHM */
+
+  if (!must_use_x_shm)
     {
-      int   rowstride = width * (bits_per_pixel / 8);
+      int   rowstride = width * (depth / 8);
       char *data;
 
       data = malloc (rowstride * height);
       if (!data)
 	return FALSE;
 
-      *image = XCreateImage (xdisplay,
-			     DefaultVisual (xdisplay, 0),
-			     bits_per_pixel,
+      *image = XCreateImage (vfb->priv->xdisplay,
+			     DefaultVisual (vfb->priv->xdisplay, 0),
+			     depth,
 			     ZPixmap,
 			     0,
 			     data,
@@ -275,17 +303,21 @@ vino_fb_create_image (VinoFB           *vfb,
 
       return FALSE;
     }
-}
 
+  return FALSE;
+}
 
 static gboolean
 vino_fb_copy_tile (VinoFB       *vfb,
-		   GdkRectangle *rect,
-		   int           bytes_per_pixel)
+		   GdkRectangle *rect)
 {
-  char *dest;
-  int   bytes_per_line;
-  int   i;
+  XImage *fb_image;
+  char   *src;
+  char   *dest;
+  int     bytes_per_pixel;
+  int     src_bytes_per_line;
+  int     dest_bytes_per_line;
+  int     i;
 
   if (!vino_fb_get_image (vfb,
 			  vfb->priv->root_window,
@@ -298,17 +330,21 @@ vino_fb_copy_tile (VinoFB       *vfb,
 			  rect->height))
     return FALSE;
 
-  bytes_per_line = vfb->priv->fb_image->bytes_per_line;
+  fb_image = vfb->priv->fb_image;
+
+  src_bytes_per_line  = vfb->priv->tile->bytes_per_line;
+  dest_bytes_per_line = fb_image->bytes_per_line;
+  bytes_per_pixel     = fb_image->bits_per_pixel >> 3;
+
+  src  = vfb->priv->tile->data;
+  dest = fb_image->data + rect->y * dest_bytes_per_line + rect->x * bytes_per_pixel;
 
-  dest = vfb->priv->fb_image->data + (rect->y * bytes_per_line) + (rect->x * bytes_per_pixel);
-  
   for (i = 0; i < rect->height; i++)
     {
-      memcpy (dest,
-	      vfb->priv->tile->data + i * bytes_per_pixel * rect->width,
-	      rect->width * bytes_per_pixel);
+      memcpy (dest, src, rect->width * bytes_per_pixel);
 
-      dest += bytes_per_line;
+      src  += src_bytes_per_line;
+      dest += dest_bytes_per_line;
     }
 
   return TRUE;
@@ -359,7 +395,7 @@ vino_fb_poll_scanline (VinoFB *vfb,
 
 	  dprintf (POLLING, "damage: (%d, %d) (%d x %d)\n", rect.x, rect.y, rect.width, rect.height);
 
-	  if (vino_fb_copy_tile (vfb, &rect, bytes_per_pixel))
+	  if (vino_fb_copy_tile (vfb, &rect))
 	    {
 	      if (!vfb->priv->damage_region)
 		vfb->priv->damage_region = gdk_region_rectangle (&rect);
@@ -413,31 +449,78 @@ vino_fb_poll_screen (VinoFB *vfb)
 }
 
 static void
-vino_fb_finalize_screen_data (VinoFB *vfb)
+vino_fb_finalize_xdamage (VinoFB *vfb)
 {
-  if (vfb->priv->damage_region)
-    gdk_region_destroy (vfb->priv->damage_region);
-  vfb->priv->damage_region = NULL;
+#ifdef HAVE_XDAMAGE
+  if (vfb->priv->fb_pixmap)
+    XFreePixmap (vfb->priv->xdisplay, vfb->priv->fb_pixmap);
+  vfb->priv->fb_pixmap = None;
 
-  if (vfb->priv->fb_image)
-    XDestroyImage (vfb->priv->fb_image);
-  vfb->priv->fb_image = NULL;
+  gdk_window_remove_filter (vfb->priv->root_window,
+			    (GdkFilterFunc) vino_fb_xdamage_event_filter,
+			    vfb);
+
+  if (vfb->priv->xdamage_copy_gc != None)
+    XFreeGC (vfb->priv->xdisplay, vfb->priv->xdamage_copy_gc);
+  vfb->priv->xdamage_copy_gc = None;
+
+  if (vfb->priv->xdamage != None)
+    XFixesDestroyRegion (vfb->priv->xdisplay, vfb->priv->xdamage_region);
+  vfb->priv->xdamage_region = None;
+
+  if (vfb->priv->xdamage != None)
+    XDamageDestroy (vfb->priv->xdisplay, vfb->priv->xdamage);
+  vfb->priv->xdamage = None;
+#endif
+}
+
+/* Frees the scanline and tile data */
+static void
+vino_fb_finalize_polling (VinoFB *vfb)
+{
+  if (vfb->priv->update_timeout)
+    g_source_remove (vfb->priv->update_timeout);
+  vfb->priv->update_timeout = 0;
 
   if (vfb->priv->scanline)
     vino_fb_destroy_image (vfb,
 			   vfb->priv->scanline,
 			   &vfb->priv->scanline_x_shm_info,
-			   vfb->priv->scanline_is_x_shm_segment);
+			   vfb->priv->scanline_is_x_shm_segment,
+			   TRUE);
   vfb->priv->scanline = NULL;
   
   if (vfb->priv->tile)
     vino_fb_destroy_image (vfb,
 			   vfb->priv->tile,
 			   &vfb->priv->tile_x_shm_info,
-			   vfb->priv->tile_is_x_shm_segment);
+			   vfb->priv->tile_is_x_shm_segment,
+			   TRUE);
   vfb->priv->tile = NULL;
 }
 
+static void
+vino_fb_finalize_screen_data (VinoFB *vfb)
+{
+  if (vfb->priv->damage_region)
+    gdk_region_destroy (vfb->priv->damage_region);
+  vfb->priv->damage_region = NULL;
+
+  if (vfb->priv->use_xdamage)
+    vino_fb_finalize_xdamage (vfb);
+  else
+    vino_fb_finalize_polling (vfb);
+
+  if (vfb->priv->fb_image)
+    vino_fb_destroy_image (vfb,
+			   vfb->priv->fb_image,
+			   &vfb->priv->fb_image_x_shm_info,
+			   vfb->priv->fb_image_is_x_shm_segment,
+			   TRUE);
+  vfb->priv->fb_image = NULL;
+
+}
+
 static void
 vino_fb_screen_size_changed (VinoFB    *vfb,
 			     GdkScreen *screen)
@@ -450,54 +533,142 @@ vino_fb_screen_size_changed (VinoFB    *vfb,
   emit_size_changed (vfb);
 }
 
-static void
-vino_fb_init_from_screen (VinoFB    *vfb,
-			  GdkScreen *screen)
+#ifdef HAVE_XDAMAGE
+static GdkFilterReturn
+vino_fb_xdamage_event_filter (GdkXEvent *xevent,
+			      GdkEvent  *event,
+			      VinoFB    *vfb)
 {
-  Display *xdisplay;
+  XEvent             *xev = (XEvent *) xevent;
+  XDamageNotifyEvent *notify;
+  GdkRectangle        damage;
+  int                 error;
 
-  g_return_if_fail (screen != NULL);
+  if (xev->type != vfb->priv->xdamage_notify_event)
+    return GDK_FILTER_CONTINUE;
 
-  xdisplay = GDK_DISPLAY_XDISPLAY (gdk_screen_get_display (screen));
+  notify = (XDamageNotifyEvent *) xev;
 
-  vfb->priv->screen      = screen;
-  vfb->priv->root_window = gdk_screen_get_root_window (screen);
+  damage.x      = notify->area.x;
+  damage.y      = notify->area.y;
+  damage.width  = notify->area.width;
+  damage.height = notify->area.height;
 
-  g_signal_connect_swapped (vfb->priv->screen, "size-changed",
-			    G_CALLBACK (vino_fb_screen_size_changed),
-			    vfb);
+  dprintf (POLLING, "Got DamageNotify event: %d %d %dx%d\n",
+	   damage.x, damage.y, damage.width, damage.height);
 
-  vfb->priv->fb_image =
-    XGetImage (xdisplay,
-	       GDK_WINDOW_XWINDOW (vfb->priv->root_window),
-	       0, 0,
-	       gdk_screen_get_width  (screen),
-	       gdk_screen_get_height (screen),
-	       AllPlanes,
-	       ZPixmap);
-  if (!vfb->priv->fb_image)
+  gdk_error_trap_push ();
+
+  /* Copy the damaged pixels from the server */
+  XCopyArea (vfb->priv->xdisplay,
+	     GDK_WINDOW_XWINDOW (vfb->priv->root_window),
+	     vfb->priv->fb_pixmap,
+	     vfb->priv->xdamage_copy_gc,
+	     damage.x,
+	     damage.y,
+	     damage.width,
+	     damage.height,
+	     damage.x,
+	     damage.y);
+  XSync (vfb->priv->xdisplay, False);
+
+  if ((error = gdk_error_trap_pop ()))
     {
-      g_warning (G_STRLOC ": failed to copy frame buffer contents with XGetImage");
+#ifdef G_ENABLE_DEBUG
+      char error_text [64];
+
+      XGetErrorText (vfb->priv->xdisplay, error, error_text, 63);
+
+      g_warning ("Received a '%s' X Window System error while copying damaged pixels",
+		 error_text);
+      g_warning ("Failed image = %d, %d %dx%d - screen = %dx%d",
+		 damage.x,
+		 damage.y,
+		 damage.width,
+		 damage.height,
+		 gdk_screen_get_width (vfb->priv->screen),
+		 gdk_screen_get_height (vfb->priv->screen));
+#endif
+      return GDK_FILTER_REMOVE;
+    }
+
+  /* add damage to our region */
+  if (vfb->priv->damage_region)
+    gdk_region_union_with_rect (vfb->priv->damage_region, &damage);
+  else
+    vfb->priv->damage_region = gdk_region_rectangle (&damage);
+
+  /* subtract damage from server */
+  XFixesSetRegion (vfb->priv->xdisplay, vfb->priv->xdamage_region, &notify->area, 1);
+  XDamageSubtract (vfb->priv->xdisplay,
+		   vfb->priv->xdamage,
+		   vfb->priv->xdamage_region,
+		   None);
+
+  emit_damage_notify (vfb);
+
+  return GDK_FILTER_REMOVE;
+}
+#endif /* HAVE_XDAMAGE */
+
+static void
+vino_fb_init_xdamage (VinoFB *vfb)
+{
+#ifdef HAVE_XDAMAGE
+  int       event_base, error_base;
+  int       major, minor;
+  XGCValues values;
+
+  if (!XDamageQueryExtension (vfb->priv->xdisplay, &event_base, &error_base))
+    return;
+
+  if (!XDamageQueryVersion (vfb->priv->xdisplay, &major, &minor) || major != 1)
+    return;
+
+  vfb->priv->xdamage_notify_event = event_base + XDamageNotify;
+
+  vfb->priv->xdamage = XDamageCreate (vfb->priv->xdisplay,
+				      GDK_WINDOW_XWINDOW (vfb->priv->root_window),
+				      XDamageReportRawRectangles);
+  if (vfb->priv->xdamage == None)
+    return;
+
+  vfb->priv->xdamage_region = XFixesCreateRegion (vfb->priv->xdisplay, NULL, 0);
+  if (vfb->priv->xdamage_region == None)
+    {
+      XDamageDestroy (vfb->priv->xdisplay, vfb->priv->xdamage);
+      vfb->priv->xdamage = None;
       return;
     }
 
-  dprintf (POLLING, "Initialized framebuffer contents (%p) for screen %d: %dx%d %dbpp\n",
-	   vfb->priv->fb_image,
-	   gdk_screen_get_number (vfb->priv->screen),
-	   vfb->priv->fb_image->width,
-	   vfb->priv->fb_image->height,
-	   vfb->priv->fb_image->bits_per_pixel);
+  values.subwindow_mode = IncludeInferiors;
+  vfb->priv->xdamage_copy_gc = XCreateGC (vfb->priv->xdisplay,
+					  GDK_WINDOW_XWINDOW (vfb->priv->root_window),
+					  GCSubwindowMode,
+					  &values);
 
-#ifdef HAVE_XSHM
-  vfb->priv->use_x_shm = XShmQueryExtension (xdisplay) != False;
+  gdk_x11_register_standard_event_type (gdk_screen_get_display (vfb->priv->screen),
+					event_base,
+					XDamageNumberEvents);
+  gdk_window_add_filter (vfb->priv->root_window,
+			 (GdkFilterFunc) vino_fb_xdamage_event_filter,
+			 vfb);
+
+  vfb->priv->use_xdamage = TRUE;
 #endif
+}
+
+static void
+vino_fb_init_polling (VinoFB *vfb)
+{
+  g_assert (!vfb->priv->use_xdamage);
 
   vfb->priv->scanline_is_x_shm_segment =
     vino_fb_create_image (vfb,
 			  &vfb->priv->scanline,
-			  &vfb->priv->scanline_x_shm_info,
-			  gdk_screen_get_width (screen), 1,
-			  vfb->priv->fb_image->bits_per_pixel);
+			  &vfb->priv->scanline_x_shm_info, FALSE,
+			  gdk_screen_get_width (vfb->priv->screen), 1,
+			  vfb->priv->fb_image->depth);
   if (!vfb->priv->scanline)
     {
       g_warning (G_STRLOC ": failed to initialize scanline XImage\n");
@@ -513,16 +684,17 @@ vino_fb_init_from_screen (VinoFB    *vfb,
   vfb->priv->tile_is_x_shm_segment =
     vino_fb_create_image (vfb,
 			  &vfb->priv->tile,
-			  &vfb->priv->tile_x_shm_info,
+			  &vfb->priv->tile_x_shm_info, FALSE,
 			  TILE_WIDTH, TILE_HEIGHT,
-			  vfb->priv->fb_image->bits_per_pixel);
+			  vfb->priv->fb_image->depth);
   if (!vfb->priv->tile)
     {
       g_warning (G_STRLOC ": failed to initialize tile XImage\n");
       vino_fb_destroy_image (vfb,
 			     vfb->priv->scanline,
 			     &vfb->priv->scanline_x_shm_info,
-			     vfb->priv->scanline_is_x_shm_segment);
+			     vfb->priv->scanline_is_x_shm_segment,
+			     TRUE);
       vfb->priv->scanline = NULL;
       XDestroyImage (vfb->priv->fb_image);
       vfb->priv->fb_image = NULL;
@@ -537,16 +709,101 @@ vino_fb_init_from_screen (VinoFB    *vfb,
     g_timeout_add (20, (GSourceFunc) vino_fb_poll_screen, vfb);
 }
 
+static void
+vino_fb_init_fb_image (VinoFB *vfb)
+{
+  if (vfb->priv->use_xdamage)
+      {
+	vfb->priv->fb_image_is_x_shm_segment =
+	  vino_fb_create_image (vfb,
+				&vfb->priv->fb_image,
+				&vfb->priv->fb_image_x_shm_info,
+				TRUE,
+				gdk_screen_get_width (vfb->priv->screen),
+				gdk_screen_get_height (vfb->priv->screen),
+				DefaultDepthOfScreen (GDK_SCREEN_XSCREEN (vfb->priv->screen)));
+      }
+
+  if (vfb->priv->fb_image)
+    {
+#ifdef HAVE_XSHM
+      vfb->priv->fb_pixmap = XShmCreatePixmap (vfb->priv->xdisplay,
+					       GDK_WINDOW_XWINDOW (vfb->priv->root_window),
+					       vfb->priv->fb_image->data,
+					       &vfb->priv->fb_image_x_shm_info,
+					       vfb->priv->fb_image->width,
+					       vfb->priv->fb_image->height,
+					       vfb->priv->fb_image->depth);
+#endif
+      if (vfb->priv->fb_pixmap == None)
+	{
+	  vino_fb_destroy_image (vfb,
+				 vfb->priv->fb_image,
+				 &vfb->priv->fb_image_x_shm_info,
+				 vfb->priv->fb_image_is_x_shm_segment,
+				 TRUE);
+	  vfb->priv->fb_image = NULL;
+	  vfb->priv->fb_image_is_x_shm_segment = FALSE;
+	}
+    }
+
+  if (!vfb->priv->fb_image)
+    {
+      vfb->priv->fb_image =
+	XGetImage (vfb->priv->xdisplay,
+		   GDK_WINDOW_XWINDOW (vfb->priv->root_window),
+		   0, 0,
+		   gdk_screen_get_width  (vfb->priv->screen),
+		   gdk_screen_get_height (vfb->priv->screen),
+		   AllPlanes,
+		   ZPixmap);
+    }
+}
+
+static void
+vino_fb_init_from_screen (VinoFB    *vfb,
+			  GdkScreen *screen)
+{
+  g_return_if_fail (screen != NULL);
+
+  vfb->priv->screen      = screen;
+  vfb->priv->xdisplay    = GDK_DISPLAY_XDISPLAY (gdk_screen_get_display (screen));
+  vfb->priv->root_window = gdk_screen_get_root_window (screen);
+
+#ifdef HAVE_XSHM
+  vfb->priv->use_x_shm = XShmQueryExtension (vfb->priv->xdisplay) != False;
+#endif
+
+  g_signal_connect_swapped (vfb->priv->screen, "size-changed",
+			    G_CALLBACK (vino_fb_screen_size_changed),
+			    vfb);
+
+  vino_fb_init_xdamage (vfb);
+  vino_fb_init_fb_image (vfb);
+
+  if (!vfb->priv->fb_image)
+    {
+      g_warning (G_STRLOC ": failed to initialize frame buffer XImage");
+      return;
+    }
+
+  dprintf (POLLING, "Initialized framebuffer contents (%p) for screen %d: %dx%d %dbpp\n",
+	   vfb->priv->fb_image,
+	   gdk_screen_get_number (vfb->priv->screen),
+	   vfb->priv->fb_image->width,
+	   vfb->priv->fb_image->height,
+	   vfb->priv->fb_image->depth);
+
+  if (!vfb->priv->use_xdamage)
+    vino_fb_init_polling (vfb);
+}
+
 
 static void
 vino_fb_finalize (GObject *object)
 {
   VinoFB *vfb = VINO_FB (object);
   
-  if (vfb->priv->update_timeout)
-    g_source_remove (vfb->priv->update_timeout);
-  vfb->priv->update_timeout = 0;
-
   vino_fb_finalize_screen_data (vfb);
  
   g_free (vfb->priv);
@@ -600,6 +857,9 @@ vino_fb_instance_init (VinoFB *vfb)
   vfb->priv = g_new0 (VinoFBPrivate, 1);
 
 #ifdef HAVE_XSHM
+  vfb->priv->fb_image_x_shm_info.shmid   = -1;
+  vfb->priv->fb_image_x_shm_info.shmaddr = (char *) -1;
+
   vfb->priv->scanline_x_shm_info.shmid   = -1;
   vfb->priv->scanline_x_shm_info.shmaddr = (char *) -1;
   
